/**
 * refcounter.h
 */
#ifndef __REFCOUNTER_INCLUDED
#define __REFCOUNTER_INCLUDED

#include <memory/export.hdr>

namespace memory {

/**
 * RefCounter -- the base class for all RefCounter-based components. ;)
 * Purpose: reference counting and automatic object removing.
 *
 * 1. This class should be ALWAYS extended as 'virtual public'. As sample,
 *
 *	class MyInterface: virtual public memory::RefCounter
 *	{
 *	public:
 *		virtual void method1() =0;
 *		virtual int method2(double, int) =0;
 *	};
 *
 * 2. Nobody should remove RefCounter-based components using 'delete' operator.
 * Use method decRefCounter() instead. As sample,
 *
 *	MyInterface *pMyInt = createMyComponent();
 *	// . . . do something
 *	pMyInt->method1();
 *	// delete pMyInt; -- this is WRONG way
 *	pMyInt->decRefCount(); // this is RIGHT way
 *	pMyInt = 0; // for sake of safety
 *
 * 3. (proposed agreement). If you want to pass RefCounter-based as parameter to
 * another function (or method), you should increment reference count like this:
 *
 *	// . . .
 *	pMyInt->incRefCount();
 *	doSomething(pMyInt);
 *	// . . .
 *
 * And if you get RefCounter-based component from parameter in your function (or method)
 * you should always decrement reference count. (do not ignore this parameter,
 * even if you do not use it).
 *
 *	void doSomething(MyInterface *pMyInt) {
 *		// do nothing with pMyInt, but remove it
 *		pMyInt->decRefCount();
 *	}
 */
class MEMORY_API RefCounter {
private:
	long m_nRefCount;
protected:
	/**
	 * virtual destructor, for overridding (reference count always equals to 0 here)
	 */
	virtual ~RefCounter();
public:
	/**
	 * constructor, initialize reference count to 1
	 */
	RefCounter();
	/**
	 * decrements reference count
	 * pointer to the object becomes unusable and would better
	 *  to be assigned to 0
	 */
	virtual void decRefCount();
	/**
	 * increments reference count
	 * this method should be used by invoker!
	 */
	virtual void incRefCount();
	virtual long _getRefCount() { return m_nRefCount; }
};

/**
 * Ptr<Class: virtual public RefCounter> -- smart pointer for RefCounter-based components
 */
template<class Class /* : virtual public RefCounter */>
class Ptr {
private:
	Class *m_pObject;
public:
	/** default contructor */
	Ptr(): m_pObject(0) {
		// do nothing more
	}
	/** initializing construtor */
	Ptr(Class *pObject): m_pObject(pObject) {
		// ref count should be always had already incremented
	}
	/** copy constructor */
	Ptr(Ptr<Class> const &o): m_pObject(o.m_pObject) {
		if(m_pObject) m_pObject->incRefCount();
	}
	/** destructor, there is not reason to be virtual, because Ptr<> is for direct use, not for extending */
	~Ptr() {
		if(m_pObject)
			m_pObject->decRefCount();
	}
	/** assignment operator */
	Ptr<Class> &operator=(Ptr<Class> const &o) {
		// do increment first
		if(o.m_pObject) o.m_pObject->incRefCount();
		if(m_pObject) m_pObject->decRefCount();
		m_pObject = o.m_pObject;
		return *this;
	}
	/** assignment operator */
	Ptr<Class> &operator=(Class *pObject) {
		if(m_pObject != pObject) { // avoid cyclic assignment
			if(m_pObject) m_pObject->decRefCount();
			m_pObject = pObject;
		}
		return *this;
	}
	/** type cast operator, increments ref count */
	operator Class *() {
		if(m_pObject)
			m_pObject->incRefCount();
		return m_pObject;
	}
	/** operator->, does not increment ref count */
	Class *operator->() {
		// do not increment
		return m_pObject;
	}
	/** operator int */
	operator int() const {
		return (m_pObject != 0);
	}
	/** operator ! */
	int operator!() const {
		return (m_pObject == 0);
	}
};

} // namespace memory

#endif
//=============================================================================
