/**
 * lockquery.cpp
 */
#include <await/await.h>
#include <await/monitor.h>
#include <await/lockmanager.h>

namespace await_n {

//-----------------------------------------------------------------------------

class SynchObject {
private:
	Monitor *m_pMonitor;
	bool m_bCont;
public:
	SynchObject(): m_pMonitor(getGlobalMonitor()), m_bCont(true) {
		m_pMonitor->enter();
	}
	~SynchObject() {
		m_pMonitor->leave();
	}
	bool cont() { return m_bCont; }
	void next() { m_bCont = false; }
};

SynchHandle::SynchHandle(): m_pObject(new SynchObject()) {}

SynchHandle::~SynchHandle() { delete m_pObject; }

bool SynchHandle::cont() { return m_pObject->cont(); }

void SynchHandle::next() { m_pObject->next(); }

//-----------------------------------------------------------------------------

class AwaitSwitchObject {
private:
	Monitor *m_pMonitor;
	AbsoluteMoment m_tMoment;
	bool m_bTimedout;
	bool m_bFired;
public:
	AwaitSwitchObject(long lTimeout): m_pMonitor(getGlobalMonitor()), m_tMoment(cvtRelativeToAbsolute(lTimeout)), m_bTimedout(false), m_bFired(false) {
		m_pMonitor->enter();
	}
	~AwaitSwitchObject() {
		m_pMonitor->awake();
		m_pMonitor->leave();
	}
	bool cont() {
		return !m_bFired;
	}
	void next() {
		if(m_bTimedout)
			m_bFired = true;
		if(!m_bFired)
			m_bTimedout = m_pMonitor->sleep(m_tMoment);
	}
	bool check(bool bFired) {
		m_bFired = bFired;
		return m_bFired;
	}
	bool may() {
		return (!m_bFired) && (!m_bTimedout);
	}
	bool timedout() {
		return m_bTimedout;
	}
};

AwaitSwitchHandle::AwaitSwitchHandle(long lTimeout): m_pObject(new AwaitSwitchObject(lTimeout)) {}

AwaitSwitchHandle::~AwaitSwitchHandle() { delete m_pObject; }

bool AwaitSwitchHandle::cont() { return m_pObject->cont(); }

void AwaitSwitchHandle::next() { m_pObject->next(); }

bool AwaitSwitchHandle::check(bool bFired) { return m_pObject->check(bFired); }

bool AwaitSwitchHandle::may() { return m_pObject->may(); }

bool AwaitSwitchHandle::timedout() { return m_pObject->timedout(); }

//-----------------------------------------------------------------------------

class LockSwitchObject {
private:
	LockManager *m_pLockManager;
	Lock *m_pLock;
	long m_lTimeout;
	ExtraInfo m_oExtraInfo;
	bool m_bTimedout;
	bool m_bVictimed;
	LockQuery m_oLockQuery;
	int m_nIndex;
	int m_nPhase;
public:
	LockSwitchObject(long lTimeout, ExtraInfo const &rInfo):
		m_pLockManager(getLockManager()), m_pLock(0), m_lTimeout(lTimeout), m_oExtraInfo(rInfo),
		m_bTimedout(false), m_bVictimed(false), m_oLockQuery(), m_nIndex(0), m_nPhase(0)
	{}
	~LockSwitchObject()	{
		if(m_pLock)
			m_pLock->unlock();
	}
	bool cont() {
		return (m_nPhase<2);
	}
	void next() {
		++m_nPhase;
		m_nIndex = 0; // rewind
		if(1 == m_nPhase) {
			try {
				m_pLock = m_pLockManager->lock(m_oLockQuery, m_lTimeout, m_oExtraInfo);
			}
			catch(TimeoutException const &) {
				m_bTimedout = true;
			}
			catch(DeadlockException const &) {
				m_bVictimed = true;
			}
			catch(LockException const &) { // unknown, assume deadlock
				m_bVictimed = true;
			}
		}
	}
	bool prep() {
		return (0 == m_nPhase);
	}
	void collect(LockSlot const &slot) {
		m_oLockQuery |= tagLock(m_nIndex++, slot);
	}
	bool selected() {
		return (m_pLock != 0) && (m_nIndex++ == m_pLock->tag());
	}
	bool timedout() {
		return m_bTimedout;
	}
	bool victimed() {
		return m_bVictimed;
	}
};

LockSwitchHandle::LockSwitchHandle(long lTimeout, ExtraInfo const &rInfo): m_pObject(new LockSwitchObject(lTimeout, rInfo)) {}

LockSwitchHandle::~LockSwitchHandle() { delete m_pObject; }

bool LockSwitchHandle::cont() { return m_pObject->cont(); }

void LockSwitchHandle::next() { m_pObject->next(); }

bool LockSwitchHandle::prep() { return m_pObject->prep(); }

void LockSwitchHandle::collect(LockSlot const &o) { m_pObject->collect(o); }

bool LockSwitchHandle::selected() { return m_pObject->selected(); }

bool LockSwitchHandle::timedout() { return m_pObject->timedout(); }

bool LockSwitchHandle::victimed() { return m_pObject->victimed(); }

} // namespace await_n

//=============================================================================
