/**
 * lockmanager.cpp
 */
#include <assert.h>
#include <stdio.h>
#include "i_lock.h"
#include "i_monitor.h"

#define CHECK_VALIDITY(gg, ll) \
	assert(gg.m_nAnyLockCount >= 0); \
	assert(gg.m_nExclusiveLockCount >= 0); \
	assert(gg.m_nAnyLockCount >= gg.m_nExclusiveLockCount); \
	assert(ll.m_nAnyLockCount >= 0); \
	assert(ll.m_nExclusiveLockCount >= 0); \
	assert(ll.m_nAnyLockCount >= ll.m_nExclusiveLockCount); \
	assert(gg.m_nAnyLockCount >= ll.m_nAnyLockCount); \
	assert(gg.m_nExclusiveLockCount >= ll.m_nExclusiveLockCount);

namespace await_n {

using namespace std;

class LockManagerImpl;
class LockImpl: public Lock {
private:
	LockManagerImpl *m_pLockManager;
	long m_nTag;
	UnitedLockTable m_LockTable;
	threadid_t m_owner;
public:
	LockImpl(LockManagerImpl *pLockManager, long nTag, UnitedLockTable const &rLockTable, threadid_t owner):
		m_pLockManager(pLockManager), m_nTag(nTag), m_LockTable(rLockTable), m_owner(owner)
	{}
	virtual ~LockImpl() {}
	// from Lock
	virtual long tag() { return m_nTag; }
	virtual void unlock();
};

struct AcquiredLockInfo {
	long m_nAnyLockCount;
	long m_nExclusiveLockCount;
	AcquiredLockInfo(): m_nAnyLockCount(0), m_nExclusiveLockCount(0) {}
};
typedef map<void const *, AcquiredLockInfo> ResourceTable;
typedef map<threadid_t, ResourceTable> ThreadTable;
typedef map<threadid_t, ChoiceLockTable> QueryTable;

typedef map<threadid_t, ExtraInfo const *> ExtraTable;

class LockManagerImpl: public LockManager {
private:
	MonitorImpl m_oMonitor; // own independent monitor
	ResourceTable m_mResourceTable;
	ThreadTable m_mThreadTable;
	QueryTable m_mQueryTable;
	ExtraTable m_mExtraTable;
	void _dumpLockTables();
public:
	LockManagerImpl();
	virtual ~LockManagerImpl();
	// from LockManager
	virtual Lock *lock(LockQuery const &, long secRelTimeout, ExtraInfo const &) throw(LockException);
	virtual void dumpLockTables();
	// own (for LockImpl)
	virtual void unlock(long tag, UnitedLockTable const &, threadid_t);
	int _may(threadid_t, long &);
	int _ddcheck(threadid_t, threadid_t) const;
};

//-----------------------------------------------------------------------------

void LockImpl::unlock() {
	assert(m_owner == getSelfThreadId());
	m_pLockManager->unlock(m_nTag, m_LockTable, m_owner);
	delete this;
}

//-----------------------------------------------------------------------------

LockManagerImpl::LockManagerImpl() {
	// do nothing more
}

LockManagerImpl::~LockManagerImpl() {
	// do nothing
}

Lock *LockManagerImpl::lock(LockQuery const &query, long sRelTimeout, ExtraInfo const &extra) throw(LockException) {
	Lock *result = 0;
	threadid_t self = getSelfThreadId();
	AbsoluteMoment sAbsTimeout = cvtRelativeToAbsolute(sRelTimeout);
	int bTimedout = 0;
	int bDeadlocked = 0;

	m_oMonitor.enter();
	{
		m_mQueryTable[self] = query.getObject()->getChoiceLockTable();
		m_mExtraTable[self] = &extra;
		long tag;
		if(!_may(self, tag)) {
			bDeadlocked = _ddcheck(self, self);
			if(!bDeadlocked) {
				do {
					bTimedout = m_oMonitor.sleep(sAbsTimeout);
					if(bTimedout) break; // do not check for ability to lock
				} while(!_may(self, tag));
			}
			else {
				fprintf(stderr, "DEADLOCK HAS BEEN DETECTED!\n");
				fprintf(stderr, "victim is %.8lx\n", self);
				_dumpLockTables();
			}
		}
		ChoiceLockTable &choice = m_mQueryTable[self];
		ResourceTable &global = m_mResourceTable;
		ResourceTable &local = m_mThreadTable[self];
		if(!bTimedout && !bDeadlocked) {
			UnitedLockTable &united = choice[tag];
			for(UnitedLockTable::const_iterator j = united.begin(), m = united.end(); j != m; ++j) {
				void const *pResource = j->first;
				int nLockType = j->second;
				AcquiredLockInfo &gg = global[pResource];
				AcquiredLockInfo &ll = local[pResource];
				gg.m_nAnyLockCount++;
				ll.m_nAnyLockCount++;
				if(nLockType == EXCLUSIVE_LOCK) {
					gg.m_nExclusiveLockCount++;
					ll.m_nExclusiveLockCount++;
				}
				CHECK_VALIDITY(gg, ll);
			}
			result = new LockImpl(this, tag, united, self);
		}
		if(local.empty())
			m_mThreadTable.erase(self);
		m_mQueryTable.erase(self);
		m_mExtraTable.erase(self);
	}
	// no need to awake threads because resource acquiring cannot cause acquiring by another thread
	m_oMonitor.leave();
	if(bTimedout)
		throw TimeoutException();
	if(bDeadlocked)
		throw DeadlockException();
	return result;
}

void LockManagerImpl::unlock(long tag, UnitedLockTable const &united, threadid_t self) {
	m_oMonitor.enter();
	{
		ResourceTable &global = m_mResourceTable;
		ResourceTable &local = m_mThreadTable[self];
		for(UnitedLockTable::const_iterator j = united.begin(), m = united.end(); j != m; ++j) {
			void const *pResource = j->first;
			int nLockType = j->second;
			AcquiredLockInfo &gg = global[pResource];
			AcquiredLockInfo &ll = local[pResource];
			gg.m_nAnyLockCount--;
			ll.m_nAnyLockCount--;
			if(nLockType == EXCLUSIVE_LOCK) {
				gg.m_nExclusiveLockCount--;
				ll.m_nExclusiveLockCount--;
			}
			CHECK_VALIDITY(gg, ll);
			if(ll.m_nAnyLockCount == 0)
				local.erase(pResource);
			if(gg.m_nAnyLockCount == 0)
				global.erase(pResource);
		}
		if(local.empty())
			m_mThreadTable.erase(self);
	}
	m_oMonitor.awake();
	m_oMonitor.leave();
}

int LockManagerImpl::_may(threadid_t self, long &tag) {
	int result = 0;
	ChoiceLockTable &choice = m_mQueryTable[self];
	ResourceTable &global = m_mResourceTable;
	ResourceTable &local = m_mThreadTable[self];
	for(ChoiceLockTable::const_iterator i = choice.begin(), n = choice.end(); i != n; ++i) {
		long ctag = i->first;
		UnitedLockTable const &united = i->second;
		int found = 1;
		for(UnitedLockTable::const_iterator j = united.begin(), m = united.end(); j != m; ++j) {
			void const *pResource = j->first;
			int nLockType = j->second;
			AcquiredLockInfo gg, ll;
			{
				ResourceTable::iterator ri;
				ri = global.find(pResource);
				if(ri != global.end())
					gg = ri->second;
				ri = local.find(pResource);
				if(ri != local.end())
					ll = ri->second;
			}
			CHECK_VALIDITY(gg, ll);
			found = (nLockType == EXCLUSIVE_LOCK)?
				(gg.m_nAnyLockCount == ll.m_nAnyLockCount):
				(gg.m_nExclusiveLockCount == ll.m_nExclusiveLockCount);
			if(!found)
				break;
		}
		if(found) {
			tag = ctag;
			result = 1;
			break;
		}
	}
	return result;
}

int LockManagerImpl::_ddcheck(threadid_t self, threadid_t another) const {
	int result = 0; // assume that another is running
	QueryTable::const_iterator ichoice = m_mQueryTable.find(another);
	if(ichoice != m_mQueryTable.end()) { // another is waiting because it is present in query table
		ChoiceLockTable const &choice = ichoice->second;
		result = 1; // assume that self is deadlocked (it happened if self would not have any way to choose)
		// trying to prove inverse, i.e. find way that might be choosen
		for(ChoiceLockTable::const_iterator i = choice.begin(), n = choice.end(); (result == 1) && (i != n); ++i) {
			UnitedLockTable const &united = i->second;
			result = 0; // assume that it's ok
			for(UnitedLockTable::const_iterator j = united.begin(), m = united.end(); (result == 0) && (j != m); ++j) {
				void const *pResource = j->first;
				int nLockType = j->second;
				for(ThreadTable::const_iterator k = m_mThreadTable.begin(), l = m_mThreadTable.end(); (result == 0) && (k != l); ++k) {
					threadid_t thr = k->first;
					if(thr == another) continue;
					int bThreadFound = 0;
					ResourceTable::const_iterator h = k->second.find(pResource);
					if(h != k->second.end()) {
						if(nLockType == EXCLUSIVE_LOCK)
							bThreadFound = (h->second.m_nAnyLockCount > 0);
						else
							bThreadFound = (h->second.m_nExclusiveLockCount > 0);
					}
					if(bThreadFound) {
						result = (self == thr) || _ddcheck(self, thr);
					}
				} // for(k = [m_mThreadTable.begin() .. end()])
			} // for(j = [united.begin() .. end()])
		} // for(i = [choice.begin() .. end()])
	} // if(choice != m_mQueryTable.end())
	return result;
}

void LockManagerImpl::dumpLockTables() {
	m_oMonitor.enter();
	try {
		_dumpLockTables();
	}
	catch(...) {
		m_oMonitor.leave();
		throw;
	}
	m_oMonitor.leave();
}

void LockManagerImpl::_dumpLockTables() {
		fprintf(stderr, "--- LOCK TABLES BEGIN ---\n");
		fprintf(stderr, "Resources\n");
		for(ResourceTable::const_iterator i = m_mResourceTable.begin(), n = m_mResourceTable.end(); i != n; ++i) {
			void const *p = i->first;
			AcquiredLockInfo const &info = i->second;
			fprintf(stderr, "\tptr = 0x%.8lx shared = %ld exclusive = %ld\n", (unsigned long) p, info.m_nAnyLockCount - info.m_nExclusiveLockCount, info.m_nExclusiveLockCount);
		}
		fprintf(stderr, "Owners\n");
		for(ThreadTable::const_iterator i = m_mThreadTable.begin(), n = m_mThreadTable.end(); i != n; ++i) {
			threadid_t t = i->first;
			ResourceTable const &mResourceTable = i->second;
			fprintf(stderr, "\tthread = 0x%.8lx\n", t);
			for(ResourceTable::const_iterator j = mResourceTable.begin(), m = mResourceTable.end(); j != m; ++j) {
				void const *p = j->first;
				AcquiredLockInfo const &info = j->second;
				fprintf(stderr, "\t\tptr = 0x%.8lx shared = %ld exclusive = %ld\n", (unsigned long) p, info.m_nAnyLockCount - info.m_nExclusiveLockCount, info.m_nExclusiveLockCount);
			}
		}
		fprintf(stderr, "Waiters\n");
		for(QueryTable::const_iterator i = m_mQueryTable.begin(), n = m_mQueryTable.end(); i != n; ++i) {
			threadid_t t = i->first;
			ChoiceLockTable const &mChoice = i->second;
			fprintf(stderr, "\tthread = 0x%.8lx", t);
			ExtraTable::iterator ei = m_mExtraTable.find(t);
			if(ei != m_mExtraTable.end())
				fprintf(stderr, " at %s:%ld %s", ei->second->m_szFilename, ei->second->m_nLine, ei->second->m_szFunction);
			fprintf(stderr, "\n");
			for(ChoiceLockTable::const_iterator j = mChoice.begin(), m = mChoice.end(); j != m; ++j) {
				long tag = j->first;
				UnitedLockTable const &mUnited = j->second;
				fprintf(stderr, "\t\ttag = %ld\n", tag);
				for(UnitedLockTable::const_iterator k = mUnited.begin(), l = mUnited.end(); k != l; ++k) {
					void const *p = k->first;
					int type = k->second;
					fprintf(stderr, "\t\t\tptr = 0x%.8lx type = %s\n", (unsigned long) p, ((type == EXCLUSIVE_LOCK)? "exclusive": "shared"));
				}
			}
		}
		fprintf(stderr, "---- LOCK TABLES END ----\n");
}

LockManager *getLockManager() {
	static LockManagerImpl g_oLockManagerImpl;
	return &g_oLockManagerImpl;
}

} // namespace await_n

//=============================================================================
