/**
 * posix/monitor.cpp
 */
#include "i_monitor.h"

namespace await_n {

enum { MAX_ABSOLUTE = 0x7fffffffffffffffLL };

MonitorImpl::MonitorImpl(): m_count(0) {
	pthread_mutexattr_t attr;
	pthread_mutexattr_init(&attr);
	pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
	pthread_mutex_init(&m_mutex, &attr);
	pthread_mutexattr_destroy(&attr);
	pthread_cond_init(&m_cond, 0);
}

MonitorImpl::~MonitorImpl() {
	pthread_cond_destroy(&m_cond);
	pthread_mutex_destroy(&m_mutex);
}

void MonitorImpl::enter() {
	pthread_mutex_lock(&m_mutex);
	if(m_count>0) // unlock if it had been already locked
		pthread_mutex_unlock(&m_mutex);
	m_count++;
}

void MonitorImpl::leave() {
	m_count--;
	if(m_count == 0)
		pthread_mutex_unlock(&m_mutex);
}

bool MonitorImpl::sleep(AbsoluteMoment moment) {
	timespec ts;
	ts.tv_nsec = (moment%1000)*1000000;
	ts.tv_sec = moment/1000;
	long tmpcount = m_count; // save lock count in thread specific memory
	m_count = 0;
	int res = pthread_cond_timedwait(&m_cond, &m_mutex, &ts);
	m_count = tmpcount; // restore lock count for current thread
	return (res == ETIMEDOUT);
}

void MonitorImpl::awake() {
	pthread_cond_broadcast(&m_cond);
}

//----------------------------------------------------------------------------

Monitor *getGlobalMonitor() {
	static MonitorImpl g_oRawMonitor;
	return &g_oRawMonitor;
}

AbsoluteMoment cvtRelativeToAbsolute(long s) {
	AbsoluteMoment result = s;
	if(result<0)
		result = MAX_ABSOLUTE;
	else {
		timeval tv;
		gettimeofday(&tv, 0);
		result += tv.tv_sec*1000LL + tv.tv_usec/1000LL;
	}
	return result;
}

threadid_t getSelfThreadId() {
	return pthread_self();
}

} // namespace await_n

//=============================================================================
